"""2-dimensional geometry point class implementation.
"""

from typing import Any
from typing import Dict
from typing import Union

from apysc._event.custom_event_interface import CustomEventInterface
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.dictionary_structure import DictionaryStructure
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface

_Int = Union[int, Int]


class Point2D(
        VariableNameInterface, RevertInterface, DictionaryStructure,
        CustomEventInterface):
    """
    2-dimensional geometry point class.

    Examples
    --------
    >>> import apysc as ap
    >>> stage: ap.Stage = ap.Stage()
    >>> sprite: ap.Sprite = ap.Sprite()
    >>> sprite.graphics.begin_fill(color='#0af')
    >>> point_1: ap.Point2D = ap.Point2D(x=0, y=0)
    >>> polygon: ap.Polygon = sprite.graphics.draw_polygon(
    ...     points=[
    ...         point_1,
    ...         ap.Point2D(x=0, y=50),
    ...         ap.Point2D(x=50, y=25),
    ...     ])

    >>> point_1.x
    Int(0)

    >>> point_1.y
    Int(0)
    """

    _x: Int
    _y: Int

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Point2D')
    def __init__(self, x: _Int, y: _Int) -> None:
        """
        2-dimensional geometry point.

        Parameters
        ----------
        x : int or Int
            X-coordinate.
        y : int or Int
            Y-coordinate.

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> polygon: ap.Polygon = sprite.graphics.draw_polygon(
        ...     points=[
        ...         ap.Point2D(x=0, y=0),
        ...         ap.Point2D(x=0, y=50),
        ...         ap.Point2D(x=50, y=25),
        ...     ])
        """
        import apysc as ap
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        from apysc._validation.number_validation import validate_integer
        validate_integer(integer=x)
        validate_integer(integer=y)
        if isinstance(x, ap.Int):
            x_: ap.Int = x
        else:
            x_ = ap.Int(x)
        if isinstance(y, ap.Int):
            y_: ap.Int = y
        else:
            y_ = ap.Int(y)
        self._x = x_
        self._y = y_
        self.variable_name = \
            expression_variables_util.get_next_variable_name(
                type_name=var_names.POINT2D)
        self._append_constructor_expression()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Point2D')
    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap
        expression: str = (
            f'var {self.variable_name} = {{'
            f'"x": {self._x.variable_name}, '
            f'"y": {self._y.variable_name}}};')
        ap.append_js_expression(expression=expression)

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Point2D')
    def x(self) -> Int:
        """
        X-coordinate property.

        Returns
        -------
        x : Int
            X-coordinate.

        Examples
        --------
        >>> import apysc as ap
        >>> point: ap.Point2D = ap.Point2D(x=50, y=100)
        >>> point.x = ap.Int(150)
        >>> point.x
        Int(150)
        """
        import apysc as ap
        x: ap.Int = ap.Int(self._x._value)
        self._append_x_getter_expression(x=x)
        return x

    @x.setter
    def x(self, value: Int) -> None:
        """
        Update x-coordinate property.

        Parameters
        ----------
        value : Int
            X-coordinate to set.
        """
        from apysc._html.debug_mode import DebugInfo
        with DebugInfo(
                callable_='x', args=[value], kwargs={},
                module_name=__name__,
                class_name=Point2D.__name__):
            import apysc as ap
            from apysc._validation.number_validation import validate_integer
            validate_integer(integer=value)
            if isinstance(value, int):
                value = ap.Int(value)
            self._x = value
            self._x._append_incremental_calc_substitution_expression()
            self._append_x_setter_expression(value=value)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Point2D')
    def _append_x_getter_expression(self, *, x: Int) -> None:
        """
        Append x property getter expression.

        Parameters
        ----------
        x : Int
            Target x value.
        """
        import apysc as ap
        expression: str = (
            f'{x.variable_name} = '
            f'{self.variable_name}["x"];'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Point2D')
    def _append_x_setter_expression(self, *, value: Int) -> None:
        """
        Append x property setter expression.

        Parameters
        ----------
        value : Int
            X-coordinate to set.
        """
        import apysc as ap
        expression: str = (
            f'{self.variable_name}["x"] = {value.variable_name};'
        )
        ap.append_js_expression(expression=expression)

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Point2D')
    def y(self) -> Int:
        """
        Y-coordinate property.

        Returns
        -------
        y : Int
            Y-coordinate.

        Examples
        --------
        >>> import apysc as ap
        >>> point: ap.Point2D = ap.Point2D(x=50, y=100)
        >>> point.y = ap.Int(150)
        >>> point.y
        Int(150)
        """
        import apysc as ap
        y: ap.Int = ap.Int(self._y._value)
        self._append_y_getter_expression(y=y)
        return y

    @y.setter
    def y(self, value: Int) -> None:
        """
        Update y-coordinate property.

        Parameters
        ----------
        value : Int
            Y-coordinate to set.
        """
        from apysc._html.debug_mode import DebugInfo
        with DebugInfo(
                callable_='y', args=[value], kwargs={},
                module_name=__name__,
                class_name=Point2D.__name__):
            import apysc as ap
            from apysc._validation.number_validation import validate_integer
            validate_integer(integer=value)
            if isinstance(value, int):
                value = ap.Int(value)
            self._y = value
            self._y._append_incremental_calc_substitution_expression()
            self._append_y_setter_expression(value=value)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Point2D')
    def _append_y_getter_expression(self, *, y: Int) -> None:
        """
        Append y property getter expression.

        Parameters
        ----------
        y : Int
            Target y value.
        """
        import apysc as ap
        expression: str = (
            f'{y.variable_name} = '
            f'{self.variable_name}["y"];'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Point2D')
    def _append_y_setter_expression(self, *, value: Int) -> None:
        """
        Append y property setter expression.

        Parameters
        ----------
        value : Int
            Y-coordinate to set.
        """
        import apysc as ap
        expression: str = (
            f'{self.variable_name}["y"] = {value.variable_name};'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Point2D')
    def __eq__(self, other: Any) -> Any:
        """
        Equal comparison method.

        Parameters
        ----------
        other : Any
            The other value to compare.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        result: ap.Boolean
        if not isinstance(other, Point2D):
            result = ap.Boolean(False)
            return result
        return other.x == self.x and other.y == self.y

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Point2D')
    def __ne__(self, other: Any) -> Any:
        """
        Not equal comparison method.

        Parameters
        ----------
        other : Any
            The other value to compare.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        result: ap.Boolean = self == other
        result = result.not_
        return result

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the
        sake of debugging).

        Returns
        -------
        repr_str : str
            Type name and coordinates values are set
            (e.g., `Point2D(Int(50), Int(100))`).
        """
        x_repr: str = repr(self._x)
        y_repr: str = repr(self._y)
        repr_str: str = f'Point2D({x_repr}, {y_repr})'
        return repr_str

    _x_snapshots: Dict[str, int]
    _y_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make values' snapshots.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._set_single_snapshot_val_to_dict(
            dict_name='_x_snapshots',
            value=int(self._x._value), snapshot_name=snapshot_name)
        self._set_single_snapshot_val_to_dict(
            dict_name='_y_snapshots',
            value=int(self._y._value), snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert values if snapshots exist.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._x._value = self._x_snapshots[snapshot_name]
        self._y._value = self._y_snapshots[snapshot_name]



"""Interface class implementation for the first control x path data.
"""

from typing import Dict

from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface


class PathControlX1Interface(RevertInterface, AttrLinkingInterface):

    _control_x1: Int

    def _initialize_control_x1_if_not_initialized(self) -> None:
        """
        Initialize the _control_x1 attribute if this instance does not
        initialize it yet.
        """
        if hasattr(self, '_control_x1'):
            return
        self._control_x1 = Int(0)

        self._append_control_x1_linking_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='PathControlX1Interface')
    def _append_control_x1_linking_setting(self) -> None:
        """
        Append a control_x1 attribute to the linking setting.
        """
        self._append_applying_new_attr_val_exp(
            new_attr=self._control_x1, attr_name='control_x1')
        self._append_attr_to_linking_stack(
            attr=self._control_x1, attr_name='control_x1')

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='PathControlX1Interface')
    def control_x1(self) -> Int:
        """
        Get a first x-coordinate of the control point.

        Returns
        -------
        control_x1 : Int
            First x-coordinate of the control point.

        Examples
        --------
        >>> import apysc as ap
        >>> bezier_3d: ap.PathBezier3D = ap.PathBezier3D(
        ...     control_x1=50, control_y1=0,
        ...     control_x2=100, control_y2=0,
        ...     dest_x=150, dest_y=50)
        >>> bezier_3d.control_x1 = ap.Int(75)
        >>> bezier_3d.control_x1
        Int(75)
        """
        self._initialize_control_x1_if_not_initialized()
        return self._control_x1._copy()

    @control_x1.setter
    def control_x1(self, value: Int) -> None:
        """
        Set a first x-coordinate of the control point.

        Parameters
        ----------
        value : Int
            First x-coordinate of the control point.
        """
        from apysc._html.debug_mode import DebugInfo
        with DebugInfo(
                callable_='control_x1', args=[value], kwargs={},
                module_name=__name__,
                class_name=PathControlX1Interface.__name__):
            self._initialize_control_x1_if_not_initialized()
            self._control_x1.value = value

            self._append_control_x1_linking_setting()

    _control_x1_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_control_x1_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_control_x1_snapshots',
            value=int(self._control_x1._value), snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if a snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._initialize_control_x1_if_not_initialized()
        self._control_x1._value = self._control_x1_snapshots[snapshot_name]



"""This module is for the constants definitions of
the expression's variable names.
"""

from typing_extensions import Final

DOCUMENT: Final[str] = 'document'
BLANK_OBJECT: Final[str] = 'bo'
DISPLAY_OBJECT: Final[str] = 'do'
PARENT: Final[str] = 'parent'
GRAPHICS: Final[str] = 'g'
POINT2D: Final[str] = 'p2d'
RECTANGLE: Final[str] = 'rect'
CIRCLE: Final[str] = 'circle'
ELLIPSE: Final[str] = 'ellipse'
LINE: Final[str] = 'line'
POLYLINE: Final[str] = 'pline'
POLYGON: Final[str] = 'poly'
PATH: Final[str] = 'path'
SPRITE: Final[str] = 'sp'
ARRAY: Final[str] = 'arr'
DICTIONARY: Final[str] = 'dct'
INDEX: Final[str] = 'idx'
BOOLEAN: Final[str] = 'b'
INT: Final[str] = 'i'
NUMBER: Final[str] = 'n'
STRING: Final[str] = 's'
SNAPSHOT: Final[str] = 'snapshot'
ANY: Final[str] = 'any'
EVENT: Final[str] = 'evt'
MOUSE_EVENT: Final[str] = 'mevt'
WHEEL_EVENT: Final[str] = 'wevt'
TIMER_EVENT: Final[str] = 'tevt'
ANIMATION_EVENT: Final[str] = 'aevt'
ANIMATION_PARALLEL_EVENT: Final[str] = 'apevt'
TIMER: Final[str] = 'timer'
ANIMATION_MOVE: Final[str] = 'animmove'
ANIMATION_X: Final[str] = 'animx'
ANIMATION_Y: Final[str] = 'animy'
ANIMATION_CX: Final[str] = 'animcx'
ANIMATION_CY: Final[str] = 'animcy'
ANIMATION_WIDTH: Final[str] = 'animwidth'
ANIMATION_HEIGHT: Final[str] = 'animheight'
ANIMATION_WIDTH_FOR_ELLIPSE: Final[str] = 'animwidthforellipse'
ANIMATION_HEIGHT_FOR_ELLIPSE: Final[str] = 'animheightforellipse'
ANIMATION_FILL_ALPHA: Final[str] = 'animfillalpha'
ANIMATION_FILL_COLOR: Final[str] = 'animfillcolor'
ANIMATION_LINE_COLOR: Final[str] = 'animlinecolor'
ANIMATION_LINE_ALPHA: Final[str] = 'animlinealpha'
ANIMATION_LINE_THICKNESS: Final[str] = 'animlinethickness'
ANIMATION_RADIUS: Final[str] = 'animradius'
ANIMATION_SKEW_X: Final[str] = 'animskewx'
ANIMATION_SKEW_Y: Final[str] = 'animskewy'
ANIMATION_ROTATION_AROUND_CENTER: Final[str] = 'animrotardcenter'
ANIMATION_ROTATION_AROUND_POINT: Final[str] = 'animrotardpoint'
ANIMATION_SCALE_X_FROM_CENTER: Final[str] = 'animscalexfromcenter'
ANIMATION_SCALE_Y_FROM_CENTER: Final[str] = 'animscaleyfromcenter'
ANIMATION_SCALE_X_FROM_POINT: Final[str] = 'animscalexfrompoint'
ANIMATION_SCALE_Y_FROM_POINT: Final[str] = 'animscaleyfrompoint'
ANIMATION_PARALLEL: Final[str] = 'animparallel'



"""Implementations of the Path class.
"""

from typing import List

from apysc._display import graphics
from apysc._display.line_base import LineBase
from apysc._geom.path_data_base import PathDataBase
from apysc._html.debug_mode import add_debug_info_setting


class Path(LineBase):
    """
    The path vector graphics class.

    Examples
    --------
    >>> import apysc as ap
    >>> stage: ap.Stage = ap.Stage()
    >>> sprite: ap.Sprite = ap.Sprite()
    >>> sprite.graphics.line_style(color='#fff', thickness=3)
    >>> path: ap.Path = sprite.graphics.draw_path(
    ...     path_data_list=[
    ...         ap.PathMoveTo(x=0, y=50),
    ...         ap.PathBezier2D(
    ...             control_x=50, control_y=0,
    ...             dest_x=100, dest_y=50),
    ...     ])
    """

    _path_data_list: List[PathDataBase]

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Path')
    def __init__(
            self,
            *,
            parent: 'graphics.Graphics',
            path_data_list: List[PathDataBase]) -> None:
        """
        Create a path vector graphic.

        Parameters
        ----------
        parent : Graphics
            Graphics instance to link this graphic.
        path_data_list : list of PathDataBase
            Target path data settings, such as the ap.PathData.MoveTo.

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=3)
        >>> path: ap.Path = sprite.graphics.draw_path(
        ...     path_data_list=[
        ...         ap.PathMoveTo(x=0, y=50),
        ...         ap.PathBezier2D(
        ...             control_x=50, control_y=0,
        ...             dest_x=100, dest_y=50),
        ...     ])
        """
        from apysc._display.graphics import Graphics
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        from apysc._validation import path_validation
        path_validation.validate_path_data_list(
            path_data_list=path_data_list)
        parent_graphics: Graphics = parent
        variable_name: str = expression_variables_util.\
            get_next_variable_name(type_name=var_names.PATH)
        super(Path, self).__init__(
            parent=parent, x=0, y=0, variable_name=variable_name)
        self._path_data_list = path_data_list
        self._set_initial_basic_values(parent=parent)
        self._append_constructor_expression()
        self._set_line_setting_if_not_none_value_exists(
            parent_graphics=parent_graphics)
        self._set_overflow_visible_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Path')
    def _append_constructor_expression(self) -> None:
        """
        Append a constructor expression.
        """
        import apysc as ap
        from apysc._display.stage import get_stage_variable_name
        from apysc._geom.path_data_util import make_paths_expression_from_list
        from apysc._string import indent_util
        stage_variable_name: str = get_stage_variable_name()
        path_data_expression: str = make_paths_expression_from_list(
            path_data_list=self._path_data_list)
        INDENT_NUM: int = 2
        expression: str = (
            f'var {self.variable_name} = {stage_variable_name}'
            f'\n  .path({path_data_expression})'
            '\n  .attr({'
        )
        expression = self._append_basic_vals_expression(
            expression=expression,
            indent_num=INDENT_NUM)
        spaces: str = indent_util.make_spaces_for_html(
            indent_num=INDENT_NUM)
        if self._fill_color._value == '':
            expression += f'\n{spaces}fill: "transparent",'
        expression += '\n  });'
        ap.append_js_expression(expression=expression)

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Path('<variable_name>')`).
        """
        repr_str: str = f"Path('{self.variable_name}')"
        return repr_str



"""Implementations of the Rectangle class.
"""

from typing import Union

from apysc._display import graphics
from apysc._display.ellipse_height_interface import EllipseHeightInterface
from apysc._display.ellipse_width_interface import EllipseWidthInterface
from apysc._display.height_interface import HeightInterface
from apysc._display.line_base import LineBase
from apysc._display.width_interface import WidthInterface
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.int import Int


class Rectangle(
        LineBase, WidthInterface, HeightInterface,
        EllipseWidthInterface, EllipseHeightInterface):
    """
    The rectangle vector graphics class.

    References
    ----------
    - Graphics draw_rect interface document
        - https://simon-ritchie.github.io/apysc/graphics_draw_rect.html

    Examples
    --------
    >>> import apysc as ap
    >>> stage: ap.Stage = ap.Stage()
    >>> sprite: ap.Sprite = ap.Sprite()
    >>> sprite.graphics.begin_fill(color='#0af')
    >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
    ...     x=50, y=50, width=100, height=75)
    >>> rectangle.x
    Int(50)

    >>> rectangle.y
    Int(50)

    >>> rectangle.width
    Int(100)

    >>> rectangle.height
    Int(75)

    >>> rectangle.fill_color
    String('#00aaff')
    """

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Rectangle')
    def __init__(
            self, *, parent: 'graphics.Graphics',
            x: Union[int, Int],
            y: Union[int, Int],
            width: Union[int, Int],
            height: Union[int, Int]) -> None:
        """
        Create a rectangle vector graphic.

        Parameters
        ----------
        parent : Graphics
            Graphics instance to link this instance.
        x : Int or int
            X-coordinate to start drawing.
        y : Int or int
            Y-coordinate to start drawing.
        width : Int or int
            Rectangle width.
        height : Int or int
            Rectangle height.

        References
        ----------
        - Graphics draw_rect interface document
            - https://simon-ritchie.github.io/apysc/graphics_draw_rect.html

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=100, height=75)
        >>> rectangle.x
        Int(50)

        >>> rectangle.y
        Int(50)

        >>> rectangle.width
        Int(100)

        >>> rectangle.height
        Int(75)

        >>> rectangle.fill_color
        String('#00aaff')
        """
        from apysc._display.graphics import Graphics
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        from apysc._validation import size_validation
        parent_graphics: Graphics = parent
        variable_name: str = expression_variables_util.\
            get_next_variable_name(type_name=var_names.RECTANGLE)
        super(Rectangle, self).__init__(
            parent=parent, x=x, y=y, variable_name=variable_name)
        size_validation.validate_size_is_gte_zero(size=width)
        size_validation.validate_size_is_gte_zero(size=height)
        self._update_width_and_skip_appending_exp(value=width)
        self._update_height_and_skip_appending_exp(value=height)
        self._set_initial_basic_values(parent=parent)
        self._append_constructor_expression()
        self._set_line_setting_if_not_none_value_exists(
            parent_graphics=parent_graphics)
        self._set_overflow_visible_setting()

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Rectangle('<variable_name>')`).
        """
        repr_str: str = f"Rectangle('{self.variable_name}')"
        return repr_str

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Rectangle')
    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap
        from apysc._display.stage import get_stage_variable_name
        variable_name: str = self.variable_name
        stage_variable_name: str = get_stage_variable_name()
        expression: str = (
            f'var {variable_name} = {stage_variable_name}'
            f'\n  .rect({self.width.variable_name}, '
            f'{self.height.variable_name})'
        )
        attrs_str: str = self._make_rect_attrs_expression()
        expression += f'{attrs_str};'
        ap.append_js_expression(expression=expression)

    def _make_rect_attrs_expression(self) -> str:
        """
        Make rectangle attributes expression string.

        Returns
        -------
        rect_attrs_expression : str
            Rectangle attributes expression string.
        """
        from apysc._display import graphics_expression
        from apysc._display.graphics import Graphics
        graphics: Graphics = self.parent_graphics
        INDENT_NUM: int = 2
        rect_attrs_expression: str = '\n  .attr({'
        rect_attrs_expression = graphics_expression.append_fill_expression(
            graphics=graphics, expression=rect_attrs_expression,
            indent_num=INDENT_NUM)
        rect_attrs_expression = \
            graphics_expression.append_fill_opacity_expression(
                graphics=graphics, expression=rect_attrs_expression,
                indent_num=INDENT_NUM)
        rect_attrs_expression = graphics_expression.append_stroke_expression(
            graphics=graphics, expression=rect_attrs_expression,
            indent_num=INDENT_NUM)
        rect_attrs_expression = \
            graphics_expression.append_stroke_width_expression(
                graphics=graphics, expression=rect_attrs_expression,
                indent_num=INDENT_NUM)
        rect_attrs_expression = \
            graphics_expression.append_stroke_opacity_expression(
                graphics=graphics, expression=rect_attrs_expression,
                indent_num=INDENT_NUM)
        rect_attrs_expression = graphics_expression.append_x_expression(
            graphic=self, expression=rect_attrs_expression,
            indent_num=INDENT_NUM)
        rect_attrs_expression = graphics_expression.append_y_expression(
            graphic=self, expression=rect_attrs_expression,
            indent_num=INDENT_NUM)

        rect_attrs_expression += '\n  })'
        return rect_attrs_expression



"""This module is for adding document translations mapping
dictionary's blank data.
"""

import os
import re
from typing import Dict
from typing import List
from typing import Match
from typing import Optional
from typing import Pattern
from typing import Union

from apysc._lint_and_doc import document_text_split_util
from apysc._lint_and_doc import lint_and_doc_hash_util
from apysc._lint_and_doc.docs_lang import Lang
from apysc._lint_and_doc.document_text_split_util import BodyText
from apysc._lint_and_doc.document_text_split_util import CodeBlock
from apysc._lint_and_doc.document_text_split_util import Heading
from apysc._lint_and_doc.lint_and_doc_hash_util import HashType
from apysc._lint_and_doc.translation_mapping_utils import MAPPING_CONST_NAME
from apysc._lint_and_doc.translation_mapping_utils import \
    MAPPING_UNNECESSARY_PATTERNS
from apysc._lint_and_doc.translation_mapping_utils import SKIPPING_PATTERNS
from apysc._lint_and_doc.translation_mapping_utils import \
    convert_splitted_values_to_keys
from apysc._lint_and_doc.translation_mapping_utils import escape_key_or_value
from apysc._lint_and_doc.translation_mapping_utils import \
    get_mapping_module_path
from apysc._lint_and_doc.translation_mapping_utils import \
    get_translated_file_path_from_src_path
from apysc._lint_and_doc.translation_mapping_utils import read_mapping_data
from apysc._lint_and_doc.translation_mapping_utils import \
    remove_escaping_from_key_or_value

_SplittedVals = List[Union[Heading, BodyText, CodeBlock]]


def add_mapping_blank_data(*, lang: Lang) -> None:
    """
    Add a mapping's blank data of document translations.

    Parameters
    ----------
    lang : Lang
        A target translation language.
    """
    from apysc._file import file_util
    src_docs_file_paths: List[str] = _get_src_docs_file_paths()
    hash_type: HashType = _get_hash_type_from_lang(lang=lang)
    src_docs_file_paths = lint_and_doc_hash_util.\
        remove_not_updated_file_paths(
            file_paths=src_docs_file_paths,
            hash_type=hash_type)
    for src_doc_file_path in src_docs_file_paths:
        markdown_txt: str = file_util.read_txt(file_path=src_doc_file_path)
        splitted_values: _SplittedVals = document_text_split_util.\
            split_markdown_document(markdown_txt=markdown_txt)
        keys: List[str] = convert_splitted_values_to_keys(
            splitted_values=splitted_values)
        keys = _remove_skipping_pattern_keys_from_list(keys=keys)
        mappings: List[Dict[str, str]] = _make_mappings_from_keys(
            keys=keys, src_doc_file_path=src_doc_file_path,
            lang=lang)
        _save_mapping_data(
            mappings=mappings, src_doc_file_path=src_doc_file_path,
            lang=lang)
    lint_and_doc_hash_util.save_target_files_hash(
        file_paths=src_docs_file_paths, hash_type=hash_type)


def _get_hash_type_from_lang(*, lang: Lang) -> HashType:
    """
    Get a hash type from a specified language type.

    Parameters
    ----------
    lang : Lang
        A target translation language.

    Returns
    -------
    hash_type : HashType
        A target hash type.

    Raises
    ------
    ValueError
        If there is no implementation of a specified language
        type's branch condition.
    """
    if lang == Lang.JP:
        return HashType.TRANSLATION_MAPPING_JP
    raise ValueError(
        'There is no implementation of a specified language type\'s '
        'branch condition. Please add a necessary branch condition: '
        f'{lang}')


def _save_mapping_data(
        *, mappings: List[Dict[str, str]],
        src_doc_file_path: str,
        lang: Lang) -> None:
    """
    Save mapping data module with specified mappings data.

    Parameters
    ----------
    mappings : list of dict
        Target mappings data to save.
    src_doc_file_path : str
        A target source document's file path.
    lang : Lang
        A target translation language.
    """
    from apysc._file import file_util
    basename: str = os.path.basename(src_doc_file_path)
    module_str: str = (
        f'"""This module is for the translation mapping data '
        'of the\nfollowing document:'
        f'\n\nDocument file: {basename}'
        f'\nLanguage: {lang.value}'
        '\n"""'
        '\n\nfrom typing import Dict'
        f'\n\n{MAPPING_CONST_NAME}: Dict[str, str] = {{'
    )
    for mapping in mappings:
        key: str = list(mapping.keys())[0]
        value: str = list(mapping.values())[0]
        module_str += f"\n\n    '{key}':"
        if len(key) >= 70:
            module_str += '  # noqa'
        module_str += f"\n    '{value}',"
        if len(value) >= 70:
            module_str += '  # noqa'
    module_str += '\n\n}\n'
    module_path: str = get_mapping_module_path(
        src_doc_file_path=src_doc_file_path, lang=lang)
    file_util.save_plain_txt(txt=module_str, file_path=module_path)


def _make_mappings_from_keys(
        *, keys: List[str],
        src_doc_file_path: str,
        lang: Lang) -> List[Dict[str, str]]:
    """
    Make mapping dictionary values from specified key values.

    Parameters
    ----------
    keys : list of str
        Target dictionary key values.
    src_doc_file_path : str
        A target source document file path.
    lang : Lang
        A target translation language.

    Returns
    -------
    mappings : list of dict
        Created mapping dictionary values. A dictionary
        value becomes a blank string if it is a new mapping value.
    """
    already_saved_mapping: Dict[str, str] = read_mapping_data(
        src_doc_file_path=src_doc_file_path, lang=lang)
    mappings: List[Dict[str, str]] = []
    for key in keys:
        key_: str = remove_escaping_from_key_or_value(key_or_val=key)
        value: str = already_saved_mapping.get(key_, '')
        value = escape_key_or_value(key_or_val=value)
        value = _set_fixed_translation_value_if_exists(
            key=key_, value=value)
        value = _set_same_value_if_code_block_mapping_is_blank(
            key=key, value=value)
        value = _convert_link_list_by_lang(
            key=key, value=value, lang=lang)
        mappings.append({key: value})
    return mappings


_LINK_PATTERN: Pattern = re.compile(
    pattern=(
        r'(.*?\[.+?\])'
        r'\((.+?\.md)\)'
        r'(?P<after_txt>.*?)'
    ),
    flags=re.MULTILINE)


def _convert_link_list_by_lang(
        *, key: str, value: str, lang: Lang) -> str:
    """
    Convert a link list value by a specified language if a key's
    value is a link list.

    Parameters
    ----------
    key : str
        A target key string (source English string).
    value : str
        A target value string.
    lang : Lang
        A target translation language.

    Returns
    -------
    value : str
        A converted value. This interface skips the
        conversion if a specified key's value is not
        a link list.
    """
    if not key.startswith('- '):
        return value
    match: Optional[Match] = _LINK_PATTERN.search(string=key)
    if match is None:
        return value
    value = ''
    while match is not None:
        prev_txt: str = match.group(1)
        doc_path: str = match.group(2)
        doc_path = get_translated_file_path_from_src_path(
            source_doc_path=doc_path, lang=lang)
        if value != '':
            value += '\n'
        value += f'{prev_txt}({doc_path})'

        key = _LINK_PATTERN.sub(
            repl=r'\g<after_txt>',
            string=key, count=1)
        match = _LINK_PATTERN.search(string=key)
    value = escape_key_or_value(key_or_val=value)
    return value


def _set_same_value_if_code_block_mapping_is_blank(
        *, key: str, value: str) -> str:
    """
    Set the same code-block value if a specified value
    is a blank string and a key's value is a code block.

    Parameters
    ----------
    key : str
        A target key string (source English string).
    value : str
        A target value string.

    Returns
    -------
    value : str
        A result value. If an argument value is not a blank
        string, this interface returns it directly.
        If a key's value is not a code block, this interface
        also returns an argument value directly.
    """
    if value != '':
        return value
    if not key.startswith('```'):
        return value
    return key


def _set_fixed_translation_value_if_exists(
        *, key: str, value: str) -> str:
    """
    Set a fixed translation value if its mapping setting exists.

    Parameters
    ----------
    key : str
        A target translation key.
    value : str
        A current translation value.

    Returns
    -------
    fixed_value : str
        A fixed translation value. If there is no fixed
        translation mapping setting, this interface directly
        returns an argument's value.
    """
    from apysc._lint_and_doc.fixed_translation_mapping import data_model
    fixed_value: str = data_model.get_fixed_translation_str_if_exists(
        key=key, lang=Lang.JP)
    if fixed_value == '':
        return value
    return fixed_value


def _remove_skipping_pattern_keys_from_list(
        *, keys: List[str]) -> List[str]:
    """
    Remove skipping pattern matching keys from a specified list.

    Parameters
    ----------
    keys : list of str
        A target key's list.

    Returns
    -------
    result_keys : list of str
        An after removing key's list.
    """
    result_keys: List[str] = []
    patterns: List[Pattern] = [
        *SKIPPING_PATTERNS, *MAPPING_UNNECESSARY_PATTERNS]
    for key in keys:
        is_pattern_matching: bool = False
        for pattern in patterns:
            match: Optional[Match] = pattern.search(string=key)
            if match is not None:
                is_pattern_matching = True
                break
        if is_pattern_matching:
            continue
        result_keys.append(key)
    return result_keys


def _get_src_docs_file_paths() -> List[str]:
    """
    Get source documents file paths.

    Returns
    -------
    src_docs_file_paths : list of str
        Source documents file paths.
    """
    dir_path: str = './docs_src/source/'
    file_or_dir_names: List[str] = os.listdir(dir_path)
    src_docs_file_paths: List[str] = []
    for file_or_dir_name in file_or_dir_names:
        file_or_dir_path: str = os.path.join(dir_path, file_or_dir_name)
        if not os.path.isfile(file_or_dir_path):
            continue
        if not file_or_dir_path.endswith('.md'):
            continue
        if _is_translated_document(path=file_or_dir_path):
            continue
        src_docs_file_paths.append(file_or_dir_path)
    return src_docs_file_paths


def _is_translated_document(*, path: str) -> bool:
    """
    Get a boolean indicating whether a specified document path
    is a translated document or not.

    Parameters
    ----------
    path : str
        A target document path.

    Returns
    -------
    result : bool
        If a specified document path is a translated document,
        this interface returns True.
    """
    basename: str = os.path.basename(path)
    for lang in Lang:
        if basename.startswith(lang.value):
            return True
    return False



"""This module is for the translation mapping data of the
following document:

Document file: display_object_get_and_set_css.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Display object get_css and set_css interfaces':
    '',

    'This page will explain the `DisplayObject` class `get_css` and `set_css` method interfaces.':  # noqa
    '',

    '## What interfaces are these?':
    '',

    'The `get_css` method will return a `css` string, and the `set_css` method will set the CSS setting to a `DisplayObject` instance.':  # noqa
    '',

    '## Basic usage':
    '',

    'Each interface requires the `name` argument as the CSS name. In addition, the `set_css` method interface also requires the `value` argument as the CSS value string.\n\nThe following example sets the `none` value to the `display` CSS if the current CSS value is the default (blank string, `\'\'`). Otherwise, revert the value to default (`Else` case) by the timer event (ticks every second).':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _SpriteOptions(TypedDict):\n    sprite: ap.Sprite\n\n\ndef on_timer(e: ap.TimerEvent, options: _SpriteOptions) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    sprite: ap.Sprite = options[\'sprite\']\n    display_css_val: ap.String = sprite.get_css(name=\'display\')\n    condition: ap.Boolean = display_css_val == \'none\'\n    with ap.If(condition):\n        sprite.set_css(name=\'display\', value=\'\')\n    with ap.Else():\n        sprite.set_css(name=\'display\', value=\'none\')\n\n\nap.Stage(\n    stage_width=150, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nsprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\noptions: _SpriteOptions = {\'sprite\': sprite}\ntimer: ap.Timer = ap.Timer(\n    handler=on_timer, delay=1000, options=options)\ntimer.start()\n\nap.save_overall_html(\n    dest_dir_path=\'display_object_get_and_set_css_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/display_object_get_and_set_css_basic_usage/index.html" width="150" height="150"></iframe>':  # noqa
    '',

    '## get_css API':
    '',

    '<!-- Docstring: apysc._display.css_interface.CssInterface.get_css -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `get_css(self, name:Union[str, apysc._type.string.String]) -> apysc._type.string.String`<hr>\n\n**[Interface summary]** Get a CSS value string.<hr>\n\n**[Parameters]**\n\n- `name`: str or String\n  - CSS name (e.g., \'display\').\n\n<hr>\n\n**[Returns]**\n\n- `css`: ap.String\n  - CSS value.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\', alpha=0.5)\n>>> sprite.set_css(name=\'display\', value=\'none\')\n>>> sprite.get_css(name=\'display\')\nString(\'none\')\n```':  # noqa
    '',

    '':
    '',

    '## set_css API':
    '',

    '<!-- Docstring: apysc._display.css_interface.CssInterface.set_css -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `set_css(self, name:Union[str, apysc._type.string.String], value:Union[str, apysc._type.string.String]) -> None`<hr>\n\n**[Interface summary]** Set a specified value string to the CSS.<hr>\n\n**[Parameters]**\n\n- `name`: str or String\n  - CSS name (e.g., \'display\').\n- `value`: str or String\n  - A CSS value string (e.g., \'none\').\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\', alpha=0.5)\n>>> sprite.set_css(name=\'display\', value=\'none\')\n>>> sprite.get_css(name=\'display\')\nString(\'none\')\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: index.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# apysc documentation':
    '',

    'Welcome to apysc documentation! apysc is the Python front-end library (currently developing and only works partially).':  # noqa
    '',

    '## Project links':
    '',

    '- [GitHub](https://github.com/simon-ritchie/apysc)\n  - Stargazers are very welcome!\n- [Twitter](https://twitter.com/apysc)\n  - The progress and updates are informed on Twitter. Please follow!\n- [PyPI](https://pypi.org/project/apysc/)':  # noqa
    '',

    '## Contents':
    '',

    '**Quick start guide**\n\n- [What apysc can do in its current implementation](what_apysc_can_do.md)\n- [Quick start guide](quick_start.md)\n- [Import conventions](import_conventions.md)\n\n---\n\n**Container classes**\n\nThe `Stage` is the apysc overall drawing area container, and the `Sprite` is the container class.\n\n- [Stage class](stage.md)\n- [Sprite class](sprite.md)\n\n---\n\n**Exporting**\n\nThe HTML and JavaScript exporting interfaces.\n\n- [Save overall html interface](save_overall_html.md)\n- [Display on the jupyter interface](display_on_jupyter.md)\n- [Display on the Google Colaboratory interface](display_on_colaboratory.md)\n- [Append js expression interface](append_js_expression.md)\n\n---\n\n**Child-related interfaces**\n\nThe parent class, such as the `Sprite` or `Stage` have the following interfaces:\n\n- [Add child and remove child interfaces](add_child_and_remove_child.md)\n- [Contains interface](contains.md)\n- [Num children interface](num_children.md)\n- [Get child at interface](get_child_at.md)\n\n---\n\n**apysc basic data classes**\n\n- [Why the apysc library does not use the Python built-in data type](why_apysc_doesnt_use_python_builtin_data_type.md)\n- [Funcdamental data classes common value interface](fundamental_data_classes_value_interface.md)\n- [Int and Number classes](int_and_number.md)\n- [Int and Number classes common arithmetic operations](int_and_number_arithmetic_operations.md)\n- [Int and Number classes common comparison operations](int_and_number_comparison_operations.md)\n- [String class](string.md)\n- [String class comparison operations](string_comparison_operations.md)\n- [String class addition and multiplication operations](string_addition_and_multiplication.md)\n- [Boolean class](boolean.md)\n- [Array class](array.md)\n- [Array class append and push interfaces](array_append_and_push.md)\n- [Array class extend and concat interfaces](array_extend_and_concat.md)\n- [Array class insert and insert at interfaces](array_insert_and_insert_at.md)\n- [Array class pop interface](array_pop.md)\n- [Array class remove and remove at interfaces](array_remove_and_remove_at.md)\n- [Array class sort interface](array_sort.md)\n- [Array class reverse interface](array_reverse.md)\n- [Array class slice interface](array_slice.md)\n- [Array class length interface](array_length.md)\n- [Array class join interface](array_join.md)\n- [Array class index of interface](array_index_of.md)\n- [Array class comparison interfaces](array_comparison.md)\n- [Dictionary class](dictionary.md)\n- [Dictionary class generic type settings](dictionary_generic.md)\n- [Dictionary class get interface](dictionary_get.md)\n- [Dictionary class length interface](dictionary_length.md)\n- [Point2D class](point2d.md)\n\n---\n\n**DisplayObject and GraphicsBase classes**\n\nThe `DisplayObject` class is the base class for each display object. The `GraphicsBase` class is the `DisplayObject` subclass and the base class of each graphics class, such as the `Rectangle` class.\n\n- [DisplayObject class](display_object.md)\n- [DisplayObject and GraphicsBase classes basic properties abstract](display_object_and_graphics_base_prop_abstract.md)\n- [DisplayObject class x and y interfaces](display_object_x_and_y.md)\n- [DisplayObject class parent interfaces](display_object_parent.md)\n- [DisplayObject class visible interface](display_object_visible.md)\n- [DisplayObject class get and set css interfaces](display_object_get_and_set_css.md)\n- [DisplayObject class mouse event binding interfaces](display_object_mouse_event.md)\n- [GraphicsBase class rotation around center interface](graphics_base_rotation_around_center.md)\n- [GraphicsBase class rotation around point interfaces](graphics_base_rotation_around_point.md)\n- [GraphicsBase class scale from center interfaces](graphics_base_scale_from_center.md)\n- [GraphicsBase class scale from point interfaces](graphics_base_scale_from_point.md)\n- [GraphicsBase class flip x and flip y interfaces](graphics_base_flip_interfaces.md)\n- [GraphicsBase class skew x and skew y interfaces](graphics_base_skew.md)\n\n---\n\n**Graphics class**\n\nThe `Graphics` class handles each vector graphics drawing.\n\n- [Draw interfaces abstract](draw_interfaces_abstract.md)\n- [Graphics class](graphics.md)\n- [Graphics class begin fill interface](graphics_begin_fill.md)\n- [Graphics class line style interface](graphics_line_style.md)\n- [Graphics class draw rect interface](graphics_draw_rect.md)\n- [Graphics class draw round rect interface](graphics_draw_round_rect.md)\n- [Graphics class draw circle interface](graphics_draw_circle.md)\n- [Graphics class draw ellipse interface](graphics_draw_ellipse.md)\n- [Graphics class move to and line to interfaces](graphics_move_to_and_line_to.md)\n- [Graphics class draw line interface](graphics_draw_line.md)\n- [Graphics class draw dotted line interface](graphics_draw_dotted_line.md)\n- [Graphics class draw dashed line interface](graphics_draw_dashed_line.md)\n- [Graphics class draw round dotted line interface](graphics_draw_round_dotted_line.md)\n- [Graphics class draw dash dotted line interface](graphics_draw_dash_dotted_line.md)\n- [Graphics class draw polygon interface](graphics_draw_polygon.md)\n- [Graphics class fill color interface](graphics_fill_color.md)\n- [Graphics class fill alpha interface](graphics_fill_alpha.md)\n- [Graphics class line color interface](graphics_line_color.md)\n- [Graphics class line alpha interface](graphics_line_alpha.md)\n- [Graphics class line thickness interface](graphics_line_thickness.md)\n- [Graphics class line dot setting interface](graphics_line_dot_setting.md)\n- [Graphics class line dash setting interface](graphics_line_dash_setting.md)\n- [Graphics class line round dot setting interface](graphics_line_round_dot_setting.md)\n- [Graphics class line dash dot setting interface](graphics_line_dash_dot_setting.md)\n\n---\n\n**Common event interfaces**\n\n- [About the handler options type](about_handler_options_type.md)\n- [Event class prevent default and stop propagation interfaces](event_prevent_default_and_stop_propagation.md)\n- [Bind and trigger custom event interfaces](bind_and_trigger_custom_event.md)\n\n---\n\n**MouseEvent class and mouse event binding**\n\n- [MouseEvent interfaces abstract](mouse_event_abstract.md)\n- [Basic mouse event interfaces](mouse_event_basic.md)\n- [Click interface](click.md)\n- [Double click interface](dblclick.md)\n- [Mousedown and mouseup interfaces](mousedown_and_mouseup.md)\n- [Mouseover and mouseout interfaces](mouseover_and_mouseout.md)\n- [Mousemove interface](mousemove.md)\n\n---\n\n**Branch instruction**\n\n- [If class](if.md)\n- [Elif class](elif.md)\n- [Else class](else.md)\n- [Each branch instruction class scope variables reverting setting](branch_instruction_variables_reverting_setting.md)\n- [Return class](return.md)\n\n---\n\n**Loop**\n\n- [For loop class](for.md)\n- [Continue class](continue.md)\n\n---\n\n**Timer**\n\n- [Timer class](timer.md)\n- [TimerEvent class](timer_event.md)\n- [Timer class delay setting](timer_delay.md)\n- [FPS enum](fps.md)\n- [Timer class repeat count setting](timer_repeat_count.md)\n- [Timer class start and stop interfaces](timer_start_and_stop.md)\n- [Timer class timer complete interface](timer_complete.md)\n- [Timer class reset interface](timer_reset.md)\n\n---\n\n**Animation**\n\n- [Animation interfaces abstract](animation_interfaces_abstract.md)\n- [AnimationEvent class](animation_event.md)\n- [Animation duration setting](animation_duration.md)\n- [Animation delay setting](animation_delay.md)\n- [Each animation interface return value](animation_return_value.md)\n- [AnimationBase class start interface](animation_base_start.md)\n- [AnimationBase class animation complete interface](animation_complete.md)\n- [AnimationBase class interfaces method chaining](animation_method_chaining.md)\n- [AnimationBase class target property](animation_base_target.md)\n- [Animation pause and play interfaces](animation_pause_and_play.md)\n- [Animation reset interface](animation_reset.md)\n- [Animation finish interface](animation_finish.md)\n- [Animation reverse interface](animation_reverse.md)\n- [Animation time interface](animation_time.md)\n- [Easing enum](easing_enum.md)\n- [Sequential animation setting](sequential_animation.md)\n- [Animation parallel interface](animation_parallel.md)\n- [Animation move interface](animation_move.md)\n- [Animation x interface](animation_x.md)\n- [Animation y interface](animation_y.md)\n- [Animation width and height interfaces](animation_width_and_height.md)\n- [Animation fill color interface](animation_fill_color.md)\n- [Animation fill alpha interface](animation_fill_alpha.md)\n- [Animation line color interface](animation_line_color.md)\n- [Animation line alpha interface](animation_line_alpha.md)\n- [Animation line thickness interface](animation_line_thickness.md)\n- [Animation radius interface](animation_radius.md)\n- [Animation rotation around center interface](animation_rotation_around_center.md)\n- [Animation rotation around point interface](animation_rotation_around_point.md)\n- [Animation scale x and y from center interfaces](animation_scale_x_and_y_from_center.md)\n- [Animation scale x and y from point interfaces](animation_scale_x_and_y_from_point.md)\n- [Animation skew x interface](animation_skew_x.md)\n\n---\n\n**Debugging**\n\n- [Trace function interface](trace.md)\n- [Set debug mode interface](set_debug_mode.md)\n- [Unset debug mode interface](unset_debug_mode.md)\n\n---\n\n**Testing**\n\n- [JavaScript assertion interface basic behavior](assertion_basic_behavior.md)\n- [Assert equal and assert not equal interfaces](assert_equal_and_not_equal.md)\n- [Assert true and assert false interfaces](assert_true_and_false.md)\n- [Assert arrays equal and arrays not equal interfaces](assert_arrays_equal_and_arrays_not_equal.md)\n- [Assert dicts equal and dicts not equal interfaces](assert_dicts_equal_and_dicts_not_equal.md)\n- [Assert defined and undefined interfaces](assert_defined_and_undefined.md)\n\n---':  # noqa
    '',

}



"""Class implementation of the width and height Interfaces
for the ellipse.

Notes
-----
A Subclass that inherits the normal WidthInterface and
HeightInterface can't use this interface.
"""

from typing import Dict

from apysc._animation.animation_height_for_ellipse_interface import \
    AnimationHeightForEllipseInterface
from apysc._animation.animation_width_for_ellipse_interface import \
    AnimationWidthForEllipseInterface
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface


class WidthAndHeightInterfacesForEllipse(
        AnimationWidthForEllipseInterface,
        AnimationHeightForEllipseInterface,
        RevertInterface, AttrLinkingInterface):

    _width: Int
    _height: Int

    def _initialize_width_and_height_if_not_initialized(self) -> None:
        """
        Initialize _width and _height attributes if this interface
        does not initialize these yet.
        """
        if not hasattr(self, '_width'):
            self._width = Int(0)
            self._append_width_attr_linking_setting()
        if not hasattr(self, '_height'):
            self._height = Int(0)
            self._append_height_attr_linking_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='WidthAndHeightInterfacesForEllipse')
    def _append_width_attr_linking_setting(self) -> None:
        """
        Append a width attribute linking settings.
        """
        self._append_applying_new_attr_val_exp(
            new_attr=self._width, attr_name='width')
        self._append_attr_to_linking_stack(
            attr=self._width, attr_name='width')

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='WidthAndHeightInterfacesForEllipse')
    def _append_height_attr_linking_setting(self) -> None:
        """
        Append a height attribute linking settings.
        """
        self._append_applying_new_attr_val_exp(
            new_attr=self._height, attr_name='height')
        self._append_attr_to_linking_stack(
            attr=self._height, attr_name='height')

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='WidthAndHeightInterfacesForEllipse')
    def width(self) -> Int:
        """
        Get a ellipse width value.

        Returns
        -------
        width : Int
            Ellipse width.

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> ellipse: ap.Ellipse = sprite.graphics.draw_ellipse(
        ...     x=100, y=100, width=50, height=50)
        >>> ellipse.width = ap.Int(100)
        >>> ellipse.width
        Int(100)
        """
        from apysc._type import value_util
        self._initialize_width_and_height_if_not_initialized()
        return value_util.get_copy(value=self._width)

    @width.setter
    def width(self, value: Int) -> None:
        """
        Update a ellipse width value.

        Parameters
        ----------
        value : int or Int
            Ellipse width value.
        """
        from apysc._html.debug_mode import DebugInfo
        with DebugInfo(
                callable_='width', args=[value], kwargs={},
                module_name=__name__,
                class_name=WidthAndHeightInterfacesForEllipse.__name__):
            import apysc as ap
            from apysc._validation import number_validation
            if not isinstance(value, ap.Int):
                number_validation.validate_integer(integer=value)
                value = ap.Int(value)
            self._width = value
            self._width._append_incremental_calc_substitution_expression()
            self._append_ellipse_width_and_height_update_expression()

            self._append_width_attr_linking_setting()

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='WidthAndHeightInterfacesForEllipse')
    def height(self) -> Int:
        """
        Get a ellipse height value.

        Returns
        -------
        height : Int
            Ellipse height.

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> ellipse: ap.Ellipse = sprite.graphics.draw_ellipse(
        ...     x=100, y=100, width=50, height=50)
        >>> ellipse.height = ap.Int(100)
        >>> ellipse.height
        Int(100)
        """
        from apysc._type import value_util
        self._initialize_width_and_height_if_not_initialized()
        return value_util.get_copy(value=self._height)

    @height.setter
    def height(self, value: Int) -> None:
        """
        Update a ellipse height value.

        Parameters
        ----------
        value : int or Int
            Ellipse height value.
        """
        from apysc._html.debug_mode import DebugInfo
        with DebugInfo(
                callable_='height', args=[value], kwargs={},
                module_name=__name__,
                class_name=WidthAndHeightInterfacesForEllipse.__name__):
            import apysc as ap
            from apysc._validation import number_validation
            if not isinstance(value, ap.Int):
                number_validation.validate_integer(integer=value)
                value = ap.Int(value)
            self._height = value
            self._height._append_incremental_calc_substitution_expression()
            self._append_ellipse_width_and_height_update_expression()

            self._append_height_attr_linking_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='WidthAndHeightInterfacesForEllipse')
    def _append_ellipse_width_and_height_update_expression(self) -> None:
        """
        Append an ellipse width and height updating expression.
        """
        import apysc as ap
        from apysc._type import value_util
        self._initialize_width_and_height_if_not_initialized()
        width_value_str: str = value_util.get_value_str_for_expression(
            value=self._width)
        height_value_str: str = value_util.get_value_str_for_expression(
            value=self._height)
        expression: str = (
            f'{self.variable_name}.radius('
            f'parseInt({width_value_str} / 2), '
            f'parseInt({height_value_str}) / 2);'
        )
        ap.append_js_expression(expression=expression)

    _width_snapshots: Dict[str, int]
    _height_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make the values' snapshots.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_width_and_height_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_width_snapshots',
            value=int(self._width._value), snapshot_name=snapshot_name)
        self._set_single_snapshot_val_to_dict(
            dict_name='_height_snapshots',
            value=int(self._height._value), snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert the values if the snapshots exist.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._width._value = self._width_snapshots[snapshot_name]
        self._height._value = self._height_snapshots[snapshot_name]



"""This module is for the translation mapping data of the
following document:

Document file: array_comparison.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Array class comparison interfaces':
    '',

    'This page explains the `Array` class comparison interfaces (equal and not equal comparison).':  # noqa
    '',

    '## Basic usage':
    '',

    'The `Array` value can compare with a Python built-in `list` value and `Array` value. A return value becomes the `Boolean` type.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\narr: ap.Array[int] = ap.Array([1, 3, 5])\nresult: ap.Boolean = arr == [1, 3, 5]\nassert result\n```':  # noqa
    '',

    '':
    '',

    '```py\n# runnable\nimport apysc as ap\n\narr: ap.Array[int] = ap.Array([1, 3, 5])\nother_arr: ap.Array[int] = ap.Array([1, 3, 5])\nresult: ap.Boolean = arr == other_arr\nassert result\n```':  # noqa
    '',

    'The equal comparison operator (`==`) and not equal comparison operator (`!=`) are supported:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\narr: ap.Array[int] = ap.Array([1, 3, 5])\nresult: ap.Boolean = arr != [2, 4, 6]\nassert result\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: sprite.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Sprite':
    '# Sprite',

    'This page explains the `Sprite` class.':
    'このページでは、`Sprite`クラスについて説明します。',

    '## What is the Sprite?':
    '## Spriteとは？',

    'The `Sprite` class is the container of each `DisplayObject` instance. It also has the `Graphics` class interfaces and can draw each vector graphic.':  # noqa
    '`Sprite`クラスは、各`DisplayObject`インスタンスのコンテナです。また、`Graphics`クラスのインタフェースを持ち、各ベクターグラフィックを描画することができます。',  # noqa

    '## Note for the automated addition':
    '## インスタンスの自動追加に関する注意点',

    'The `Sprite` instance is automated added to the stage (no need to call `add_child` or other related interfaces). However, suppose you want to add to the other instance. In that case, you need to call the `add_child` method manually.':  # noqa
    '`Sprite`のインスタンスは自動的にステージに追加されます（`add_child`などの関連インタフェースを呼び出す必要はありません）。一方で、もし他のインスタンスに`Sprite`のインスタンスを追加したいと場合、手動で `add_child` メソッドを呼び出す必要があります。',  # noqa

    '## graphics attribute interfaces':
    '## graphics属性のインタフェース',

    'The `Sprite` instance has the `graphics` attribute, and you can draw each vector graphic with it. For example, the following code draws the cyan color rectangle.':  # noqa
    '`Sprite`クラスのインスタンスは`graphics`属性を持っており、それを使って各ベクターグラフィックを描画することができます。例えば以下のコードでは水色の四角を描画します。',  # noqa

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nsprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\n\nap.save_overall_html(\n    dest_dir_path=\'sprite_graphics_attribute/\')\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nsprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\n\nap.save_overall_html(\n    dest_dir_path=\'sprite_graphics_attribute/\')\n```',  # noqa

    'For more details, please see the `Graphics` related documents, for example:':  # noqa
    '詳細は以下の`Graphics`クラスの関連ドキュメントをご覧ください。',

    '- [Graphics class](graphics.md)\n- [Graphics class begin fill interface](graphics_begin_fill.md)\n- [Graphics class line style interface](graphics_line_style.md)\n- [Graphics class draw rect interface](graphics_draw_rect.md)\n- [Graphics class draw circle interface](graphics_draw_circle.md)':  # noqa
    '- [Graphics class](jp_graphics.md)\n\n- [Graphics class begin fill interface](jp_graphics_begin_fill.md)\n\n- [Graphics class line style interface](jp_graphics_line_style.md)\n\n- [Graphics class draw rect interface](jp_graphics_draw_rect.md)\n\n- [Graphics class draw circle interface](jp_graphics_draw_circle.md)',  # noqa

    '## Move DisplayObject instances simultaneously':
    '## DisplayObjectの複数のインスタンスの移動について',

    'The `Sprite` class is a container, and if you move that coordinates, it changes children\'s coordinates simultaneously. For example, the following code changes the sprite y-coordinate when clicking the rectangle.':  # noqa
    '`Sprite`クラスはコンテナであり、その座標を移動させると同時に子のインスタンスの座標も変更されます。例えば以下のコードでは四角をクリックするとSpriteのy座標が変化します（子の各四角形が一通り移動します）。',  # noqa

    '```py\n# runnable\nimport apysc as ap\n\n\ndef on_sprite_click(\n        e: ap.MouseEvent[ap.Sprite], options: dict) -> None:\n    """\n    The handler that the sprite calls when clicked.\n\n    Parameters\n    ----------\n    e : MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    sprite: ap.Sprite = e.this\n    sprite.y += 50\n\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=250,\n    stage_height=250,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nsprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\nsprite.graphics.draw_rect(x=150, y=50, width=50, height=50)\nsprite.click(on_sprite_click)\n\nap.save_overall_html(\n    dest_dir_path=\'sprite_move_instances_simultaneously/\')\n```':  # noqa
    '```py\n# runnable\nimport apysc as ap\n\n\ndef on_sprite_click(\n        e: ap.MouseEvent[ap.Sprite], options: dict) -> None:\n    """\n    The handler that the sprite calls when clicked.\n\n    Parameters\n    ----------\n    e : MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    sprite: ap.Sprite = e.this\n    sprite.y += 50\n\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=250,\n    stage_height=250,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nsprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\nsprite.graphics.draw_rect(x=150, y=50, width=50, height=50)\nsprite.click(on_sprite_click)\n\nap.save_overall_html(\n    dest_dir_path=\'sprite_move_instances_simultaneously/\')\n```',  # noqa

    'The subsequent pages explain the other interfaces, such as the `add_child` interface.':  # noqa
    '以降のページでは`add_child`のインターフェースなど、Spriteクラスの他のインターフェースについて説明していきます。',

    '## See also':
    '## 関連資料',

    '- [Add child and remove child interfaces](add_child_and_remove_child.md)\n- [Contains interface](contains.md)\n- [Num children interface](num_children.md)\n- [Get child at interface](get_child_at.md)':  # noqa
    '- [Add child and remove child interfaces](jp_add_child_and_remove_child.md)\n\n- [Contains interface](jp_contains.md)\n\n- [Num children interface](jp_num_children.md)\n\n- [Get child at interface](jp_get_child_at.md)',  # noqa

    '## Sprite class constructor API':
    '## SpriteクラスのコンストラクタAPI',

    '<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>':  # noqa
    '<span class="inconspicuous-txt">特記事項: このAPIドキュメントはドキュメントビルド用のスクリプトによって自動で生成・同期されています。そのためもしかしたらこの節の内容は前節までの内容と重複している場合があります。</span>',  # noqa

    '**[Interface summary]** Create a basic display object that can be a parent.<hr>':  # noqa
    '**[インターフェイス概要]** 子を持つことのできる基本的な表示要素用のオブジェクトを生成します。',

    '**[Parameters]**':
    '**[引数]**',

    '- `variable_name`: str or None, default None\n  - Variable name of this instance. A js expression uses this setting. It is unnecessary to specify any string except when instantiating the `Sprite` subclass.':  # noqa
    '- `variable_name`: str or None, default None\n  - このインスタンスのJavaScriptで使用するための変数名です。Spriteクラスのサブクラスを使う場合を除いてこの引数の指定は不要です。',  # noqa

    '**[Examples]**':
    '**[コードサンプル]**',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite_1: ap.Sprite = ap.Sprite()\n>>> # Create the sprite child rectangle\n>>> sprite_1.graphics.begin_fill(color=\'#0af\')\n>>> rect: ap.Rectangle = sprite_1.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> sprite_1.graphics.contains(rect)\nBoolean(True)\n\n>>> # Move the created rectangle to the other sprite\n>>> sprite_2: ap.Sprite = ap.Sprite()\n>>> sprite_2.add_child(rect)\n>>> sprite_1.graphics.contains(rect)\nBoolean(False)\n\n>>> sprite_2.contains(rect)\nBoolean(True)\n\n>>> # Move the sprite container\n>>> sprite_2.x = ap.Int(50)\n>>> sprite_2.x\nInt(50)\n```':  # noqa
    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite_1: ap.Sprite = ap.Sprite()\n>>> # Create the sprite child rectangle\n>>> sprite_1.graphics.begin_fill(color=\'#0af\')\n>>> rect: ap.Rectangle = sprite_1.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> sprite_1.graphics.contains(rect)\nBoolean(True)\n\n>>> # Move the created rectangle to the other sprite\n>>> sprite_2: ap.Sprite = ap.Sprite()\n>>> sprite_2.add_child(rect)\n>>> sprite_1.graphics.contains(rect)\nBoolean(False)\n\n>>> sprite_2.contains(rect)\nBoolean(True)\n\n>>> # Move the sprite container\n>>> sprite_2.x = ap.Int(50)\n>>> sprite_2.x\nInt(50)\n```',  # noqa

}



"""Class implementation of the floating-point number value.
"""

from typing import Any
from typing import Union

from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.number_value_interface import NumberValueInterface


class Number(NumberValueInterface[float, 'Number']):
    """
    Floating point number class for the apysc library.

    Notes
    -----
    The `Float` class is the alias of the Number,
    and it behaves the same as the Number class.

    References
    ----------
    - Int and Number document
        - https://simon-ritchie.github.io/apysc/int_and_number.html
    - Int and Number common arithmetic operations document
        - https://simon-ritchie.github.io/apysc/int_and_number_arithmetic_operations.html  # noqa
    - Int and Number common comparison operations document
        - https://simon-ritchie.github.io/apysc/int_and_number_comparison_operations.html  # noqa

    Examples
    --------
    >>> import apysc as ap
    >>> number: ap.Number = ap.Number(10.5)
    >>> number
    Number(10.5)

    >>> number == 10.5
    Boolean(True)

    >>> number == ap.Number(10.5)
    Boolean(True)

    >>> number >= 10.5
    Boolean(True)

    >>> number += 10.3
    >>> number
    Number(20.8)
    """

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Number')
    def __init__(
            self,
            value: Union[int, float, NumberValueInterface]) -> None:
        """
        Floating point number class for apysc library.

        Parameters
        ----------
        value : int or float or Int or Number
            Initial floating point number value. This class
            casts it to float if you specify int or Int value.

        Notes
        -----
        The `Float` class is the alias of the Number, and it behaves
        the same as the Number class.

        References
        ----------
        - Int and Number document
            - https://simon-ritchie.github.io/apysc/int_and_number.html
        - Int and Number common arithmetic operations document
            - https://simon-ritchie.github.io/apysc/int_and_number_arithmetic_operations.html  # noqa
        - Int and Number common comparison operations document
            - https://simon-ritchie.github.io/apysc/int_and_number_comparison_operations.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> number: ap.Number = ap.Number(10.5)
        >>> number
        Number(10.5)

        >>> number == 10.5
        Boolean(True)

        >>> number == ap.Number(10.5)
        Boolean(True)

        >>> number >= 10.5
        Boolean(True)

        >>> number += 10.3
        >>> number
        Number(20.8)
        """
        from apysc._converter import cast
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        from apysc._expression.event_handler_scope import \
            TemporaryNotHandlerScope
        with TemporaryNotHandlerScope():
            TYPE_NAME: str = var_names.NUMBER
            self.variable_name = expression_variables_util.\
                get_next_variable_name(type_name=TYPE_NAME)
            super(Number, self).__init__(value=value, type_name=TYPE_NAME)
            self._value = cast.to_float_from_int(int_or_float=self.value)
            self.append_constructor_expression()

    def _set_value_and_skip_expression_appending(
            self, *, value: Union[int, float, Any]) -> None:
        """
        Update value attribute and skip expression appending.

        Parameters
        ----------
        value : int or float or Int or Number
            Any number value to set. If a float or Number
            value is specified, this interface casts its
            value to an integer.
        """
        from apysc._converter import cast
        from apysc._validation import number_validation
        number_validation.validate_num(num=value)
        if isinstance(value, NumberValueInterface):
            value._value = cast.to_float_from_int(int_or_float=value._value)
            value_ = value._value
        else:
            value = cast.to_float_from_int(int_or_float=value)
            value_ = value
        self._value = value_

    def __repr__(self) -> str:
        """
        Get a representation string of this instance.

        Returns
        -------
        repr_str : str
            Representation string of this instance.
        """
        if not hasattr(self, '_value'):
            repr_str: str = 'Number(0)'
        else:
            repr_str = f'Number({self._value})'
        return repr_str



"""Stage (canvas) implementation.

References
----------
- Stage document
    - https://simon-ritchie.github.io/apysc/stage.html
"""

import ctypes
import random
from datetime import datetime
from typing import Optional
from typing import Tuple

from apysc._display.child_interface import ChildInterface
from apysc._display.height_interface import HeightInterface
from apysc._display.width_interface import WidthInterface
from apysc._event.custom_event_interface import CustomEventInterface
from apysc._event.mouse_event_interfaces import MouseEventInterfaces
from apysc._type.variable_name_interface import VariableNameInterface

# This is used only for avoiding gabage collection.
_current_stage: 'Stage'


class Stage(
        ChildInterface, WidthInterface, HeightInterface,
        VariableNameInterface, MouseEventInterfaces,
        CustomEventInterface):
    """
    The Stage (overall viewport) class.

    References
    ----------
    Stage document
        https://simon-ritchie.github.io/apysc/stage.html

    Examples
    --------
    >>> import apysc as ap
    >>> stage: ap.Stage = ap.Stage(
    ...     stage_width=500, stage_height=300,
    ...     background_color='#333', stage_elem_id='sales_chart')
    """

    _background_color: str
    _add_to: str
    _stage_elem_id: str
    stage: 'Stage'

    def __init__(
            self, *, stage_width: int = 300, stage_height: int = 185,
            background_color: str = '#ffffff',
            add_to: str = 'body',
            stage_elem_id: Optional[str] = None) -> None:
        """
        Create Stage (overall viewport) instance.

        Parameters
        ----------
        stage_width : int, default 300
            Stage width.
        stage_height : int, default 185
            Stage height
        background_color : str, default '#ffffff'
            Hexadecimal background color string.
        add_to : str, default 'body'
            Specification of element to add stage.
            Unique tag (e.g., 'body') or ID selector
            (e.g., '#any-unique-elem') is acceptable.
        stage_elem_id : str or None, optional
            ID attribute set to stage html element (e.g., 'line-graph').
            If None is set, random integer will be applied.

        References
        ----------
        Stage document
            https://simon-ritchie.github.io/apysc/stage.html

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage(
        ...     stage_width=500, stage_height=300,
        ...     background_color='#333', stage_elem_id='sales_chart')
        """
        import apysc as ap
        from apysc._color import color_util
        from apysc._expression import expression_data_util
        from apysc._html import html_util
        from apysc._validation import string_validation
        global _current_stage
        expression_data_util.empty_expression()
        self.stage = self
        self._stage_elem_id = self._create_stage_elem_id_if_none(
            stage_elem_id=stage_elem_id)
        string_validation.validate_not_empty_string(
            string=self._stage_elem_id)
        self._save_stage_elem_id()
        self._stage_elem_id = html_util.remove_first_selector_symbol_char(
            str_val=self._stage_elem_id)
        self.variable_name = get_stage_variable_name()
        self._update_width_and_skip_appending_exp(value=ap.Int(stage_width))
        self._update_height_and_skip_appending_exp(value=ap.Int(stage_height))

        background_color = color_util.complement_hex_color(
            hex_color_code=background_color)
        self._background_color = background_color
        string_validation.validate_not_empty_string(string=add_to)
        self._add_to = add_to
        self._append_constructor_expression()
        self._children = ap.Array([])
        _save_stage_id_to_db(stage=self)
        _current_stage = self

    def _save_stage_elem_id(self) -> None:
        """
        Save the stage element id.
        """
        from apysc._expression import expression_data_util
        table_name: str = expression_data_util.TableName.STAGE_ELEM_ID.value
        query: str = f'DELETE FROM {table_name};'
        expression_data_util.exec_query(sql=query, commit=False)
        query = (
            f'INSERT INTO {table_name}(elem_id) '
            f"VALUES ('{self._stage_elem_id}');")
        expression_data_util.exec_query(sql=query)

    def _create_stage_elem_id_if_none(
            self, *, stage_elem_id: Optional[str]) -> str:
        """
        Create random stage element id if a specified id is None.

        Parameters
        ----------
        stage_elem_id : str or None
            Specified stage element id.

        Returns
        -------
        result_id : str
            If a specified id isn't None, this interface
            returns an unchanged argument value. Otherwise,
            this interface returns a random integer string.
        """
        if stage_elem_id is not None:
            return stage_elem_id
        now_timestamp: int = int(datetime.now().timestamp() * 1000)
        random_int: int = random.randint(1000000, 10000000)
        result_id: str = f'stage_{now_timestamp}{random_int}'
        return result_id

    def _append_constructor_expression(self) -> None:
        """
        Append stage constructor expression.
        """
        import apysc as ap
        expression: str = self._make_constructor_expression()
        ap.append_js_expression(expression=expression)

    def _make_constructor_expression(self) -> str:
        """
        Make a stage constructor expression string.

        Returns
        -------
        expression : str
            Result expression.
        """
        style: str = self._make_style_str()
        expression: str = (
            f'var stage_html = \'<div id="{self._stage_elem_id}" '
            f'style="{style}"></div>\';'
            f'\n$("{self._add_to}").append(stage_html);'
            f'\n{get_stage_variable_name()} = SVG()'
            f'.addTo("#{self._stage_elem_id}").size('
            f'\n  {self.width}, {self.height});'
        )
        return expression

    def _make_style_str(self) -> str:
        """
        Make a stage's style string.

        Returns
        -------
        style : str
            Result style string (width, height, etc).
        """
        style: str = (
            f'width: {self.width}px;'
            f' height: {self.height}px;'
            f' background-color: {self._background_color};'
        )
        return style

    @property
    def stage_elem_id(self) -> str:
        """
        Get stage's html element id.

        Returns
        -------
        stage_elem_id : str
            Stage's html element id (not including class or id symbol).
            e.g., 'line-graph'

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage(
        ...     stage_width=500, stage_height=300,
        ...     background_color='#333', stage_elem_id='sales_chart')
        >>> stage.stage_elem_id
        'sales_chart'
        """
        return self._stage_elem_id

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and elem id will be set
            (e.g., `Stage('<stage_elem_id>')`).
        """
        repr_str: str = f"Stage('{self.stage_elem_id}')"
        return repr_str


def get_stage_elem_id() -> str:
    """
    Get current stage's element id.

    Returns
    -------
    stage_elem_id : str
        Current stage's element id. If there is no
        instantiated stage yet, this interface returns
        a blank string.
    """
    from apysc._expression import expression_data_util
    table_name: str = expression_data_util.TableName.STAGE_ELEM_ID.value
    query: str = (
        f'SELECT elem_id FROM {table_name} LIMIT 1;'
    )
    expression_data_util.exec_query(sql=query)
    result: Optional[Tuple[str]] = expression_data_util.cursor.fetchone()
    if result is None:
        return ''
    return result[0]


def get_stage_variable_name() -> str:
    """
    Get current stage's global variable name.

    Returns
    -------
    stage_variable_name : str
        Current stage's js global variable name.
        If there is no instantiated stage yet, this interface
        returns a blank string.
    """
    stage_elem_id: str = get_stage_elem_id()
    stage_variable_name: str = stage_elem_id.replace('-', '_')
    return stage_variable_name


def get_stage_elem_str() -> str:
    """
    Get current stage's jQuery element string.

    Returns
    -------
    stage_elem_str : str
        Stage's jQuery element string (e.g., '$("#<stage_elem_id>")').
    """
    stage_elem_id: str = get_stage_elem_id()
    stage_elem_str: str = f'$("#{stage_elem_id}")'
    return stage_elem_str


def _save_stage_id_to_db(stage: Stage) -> None:
    """
    Save a stage's memory address (id) to the database.

    Parameters
    ----------
    stage : Stage
        Target stage.
    """
    from apysc._expression import expression_data_util
    table_name: str = expression_data_util.TableName.STAGE_ID.value
    query: str = f'DELETE FROM {table_name};'
    expression_data_util.exec_query(sql=query, commit=False)
    id_: int = id(stage)
    query = (
        f'INSERT INTO {table_name}(stage_id) '
        f'VALUES ({id_});'
    )
    expression_data_util.exec_query(sql=query)


class _StageNotCreatedError(Exception):
    pass


def get_stage() -> Stage:
    """
    Get an already instantiated stage instance.

    Returns
    -------
    stage : Stage
        Target stage instance.

    Raises
    ------
    _StageNotCreatedError
        If there is no instantiated stage yet.
    """
    from apysc._expression import expression_data_util
    table_name: str = expression_data_util.TableName.STAGE_ID.value
    query: str = f'SELECT stage_id FROM {table_name} LIMIT 1;'
    expression_data_util.exec_query(sql=query)
    result: Optional[Tuple[int]] = expression_data_util.cursor.fetchone()
    if result is None:
        raise _StageNotCreatedError(
            'Stage is not instantiated yet. Please instantiate the '
            'ap.Stage class before calling this function.')
    stage: Stage = ctypes.cast(result[0], ctypes.py_object).value
    return stage



"""Class implementation for the center x-coordinate interface.
"""

from typing import Dict

from apysc._animation.animation_cx_interface import AnimationCxInterface
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface


class CxInterface(
        AnimationCxInterface, RevertInterface, AttrLinkingInterface):

    _cx: Int

    def _initialize_cx_if_not_initialized(self) -> None:
        """
        Initialize _cx attribute if this interface does not
        initialize it yet.
        """
        if hasattr(self, '_cx'):
            return
        self._cx = Int(0)

        self._append_cx_attr_linking_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='CxInterface')
    def _append_cx_attr_linking_setting(self) -> None:
        """
        Append cx attribute linking settings.
        """
        self._append_applying_new_attr_val_exp(
            new_attr=self._cx, attr_name='cx')
        self._append_attr_to_linking_stack(attr=self._cx, attr_name='cx')

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='CxInterface')
    def x(self) -> Int:
        """
        Get a center x-coordinate.

        Returns
        -------
        x : Int
            Center x-coordinate.

        References
        ----------
        - Display object x and y interfaces document
            - https://simon-ritchie.github.io/apysc/display_object_x_and_y.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af', alpha=0.5)
        >>> circle: ap.Circle = sprite.graphics.draw_circle(
        ...     x=100, y=100, radius=50)
        >>> circle.x = ap.Int(120)
        >>> circle.x
        Int(120)
        """
        import apysc as ap
        from apysc._type import value_util
        self._initialize_cx_if_not_initialized()
        x: ap.Int = value_util.get_copy(value=self._cx)
        return x

    @x.setter
    def x(self, value: Int) -> None:
        """
        Update a center x-coordinate.

        Parameters
        ----------
        value : int or Int
            Center x-coordinate value.

        References
        ----------
        - Display object x and y interfaces document
            - https://simon-ritchie.github.io/apysc/display_object_x_and_y.html  # noqa
        """
        from apysc._html.debug_mode import DebugInfo
        with DebugInfo(
                callable_='x', args=[value], kwargs={},
                module_name=__name__,
                class_name=CxInterface.__name__):
            import apysc as ap
            from apysc._validation import number_validation
            number_validation.validate_integer(integer=value)
            if not isinstance(value, ap.Int):
                value = ap.Int(value)
            self._cx = value
            self._cx._append_incremental_calc_substitution_expression()
            self._append_cx_update_expression()

            self._append_cx_attr_linking_setting()

    def _append_cx_update_expression(self) -> None:
        """
        Append cx position updating expression.
        """
        import apysc as ap
        from apysc._type import value_util
        self._initialize_cx_if_not_initialized()
        value_str: str = value_util.get_value_str_for_expression(
            value=self._cx)
        expression: str = (
            f'{self.variable_name}.cx({value_str});'
        )
        ap.append_js_expression(expression=expression)

    _cx_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_cx_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_cx_snapshots',
            value=int(self._cx._value), snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if a snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._cx._value = self._cx_snapshots[snapshot_name]



"""This module is for the translation mapping data of the
following document:

Document file: get_child_at.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# get_child_at interface':
    '',

    'This page explains the container class, like the `Graphics`\\, `Sprite`\\, `Stage`) `get_child_at` method interface.':  # noqa
    '',

    '## What interface is this?':
    '',

    'The `get_child_at` interface returns a child (`DisplayObject`) instance at the specified index.':  # noqa
    '',

    '## Basic usage':
    '',

    'The following code example is adding the rectangle to the sprite container. The `Sprite` class adds the `Graphics` instance at the constructor so that the first child becomes the `Graphics` instance. The second child becomes the `Rectangle` instance, which the sprite added with the `add_child` method.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=450,\n    stage_height=150,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle_1: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nsprite.add_child(rectangle_1)\n\nfirst_child: ap.DisplayObject = sprite.get_child_at(index=0)\nassert isinstance(first_child, ap.Graphics)\n\nsecond_child: ap.DisplayObject = sprite.get_child_at(index=1)\nassert isinstance(second_child, ap.Rectangle)\n```':  # noqa
    '',

    '':
    '',

    '## get_child_at API':
    '',

    '<!-- Docstring: apysc._display.child_interface.ChildInterface.get_child_at -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `get_child_at(self, index:Union[int, apysc._type.int.Int]) -> apysc._display.display_object.DisplayObject`<hr>\n\n**[Interface summary]** Get child at a specified index.<hr>\n\n**[Parameters]**\n\n- `index`: int or Int\n  - Child\'s index (start from 0).\n\n<hr>\n\n**[Returns]**\n\n- `child`: DisplayObject\n  - Target index child instance.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\', alpha=0.5)\n>>> rectangle_1: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> rectangle_2: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=150, y=50, width=50, height=50)\n>>> child_at_index_1: ap.DisplayObject = (\n...     sprite.graphics.get_child_at(1))\n>>> child_at_index_1 == rectangle_2\nTrue\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: graphics_line_thickness.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Graphics line_thickness interface':
    '',

    'This page explains the `Graphics` class `line_thickness` property interface.':  # noqa
    '',

    '## What interface is this?':
    '',

    'The `line_thickness` property interface updates or get the instance\'s line thickness (line width).':  # noqa
    '',

    '## Basic usage':
    '',

    'The getter or setter interface value becomes (or requires) the `Int` value.\n\nThe following example sets the 5-pixel line thickness to the first rectangle and the 10-pixel line thickness to the second one:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    stage_width=250, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.line_style(color=\'#0af\', thickness=1)\n\nrectangle_1: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nrectangle_1.line_thickness = ap.Int(5)\n\nrectangle_2: ap.Rectangle = sprite.graphics.draw_rect(\n    x=150, y=50, width=50, height=50)\nrectangle_2.line_thickness = ap.Int(10)\n\nap.save_overall_html(\n    dest_dir_path=\'./graphics_line_thickness_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_line_thickness_basic_usage/index.html" width="250" height="150"></iframe>':  # noqa
    '',

    '## line_thickness property API':
    '',

    '<!-- Docstring: apysc._display.line_thickness_interface.LineThicknessInterface.line_thickness -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get this instance\'s line thickness.<hr>\n\n**[Returns]**\n\n- `line_thickness`: Int\n  - Current line thickness.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.line_style(\n...     color=\'#fff\', thickness=5)\n>>> line: ap.Line = sprite.graphics.draw_line(\n...     x_start=50, y_start=50, x_end=150, y_end=50)\n>>> line.line_thickness\nInt(5)\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: string_addition_and_multiplication.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# String class addition and multiplication operations':
    '',

    'This page explains the `String` class addition and multiplication operations.':  # noqa
    '',

    '## Addition':
    '',

    'The `String` class addition operation (`+`) returns the concatenated `String` value:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstring_1: ap.String = ap.String(\'Hello\')\nstring_2: ap.String = string_1 + \' World!\'\nassert string_2 == \'Hello World!\'\nassert isinstance(string_2, ap.String)\n```':  # noqa
    '',

    'Also, the `+=` operator is supported:':
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstring_1: ap.String = ap.String(\'Hello\')\nstring_1 += \' World!\'\nassert string_1 == \'Hello World!\'\n```':  # noqa
    '',

    'A `String` value + Python built-in `str` operation is supported. Similarly, a `String` value + `String` value operation is also supported:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstring_1: ap.String = ap.String(\'Hello\')\nstring_2: ap.String = ap.String(\' World!\')\nstring_3: ap.String = string_1 + string_2\nassert string_3 == \'Hello World!\'\n```':  # noqa
    '',

    'But a Python built-in `str` + `String` value is not supported; for instance, the following code raises an error:':  # noqa
    '',

    '```py\nimport apysc as ap\n\nstring_1: ap.String = ap.String(\' World!\')\nstring_2: ap.String = \'Hello\' + string_1\n```':  # noqa
    '',

    '':
    '',

    '```\nTypeError: must be str, not String\n```':
    '',

    '':
    '',

    '## Multiplication':
    '',

    'The `String` class multiplication operation (`*`) returns the repeated `String` value, same behaviors as the Python built-in `str` value:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstring_1: ap.String = ap.String(\'Hello\')\nstring_2: ap.String = string_1 * 3\nassert string_2 == \'HelloHelloHello\'\n```':  # noqa
    '',

    'The `int` or `Int` values are acceptable at the operation\'s right-side value:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstring_1: ap.String = ap.String(\'Hello\')\nint_1: ap.Int = ap.Int(3)\nstring_2: ap.String = string_1 * int_1\nassert string_2 == \'HelloHelloHello\'\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: graphics_draw_ellipse.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Graphics draw_ellipse interface':
    '',

    'This page explains the `Graphics` class `draw_ellipse` method interface.':  # noqa
    '',

    '## What interface is this?':
    '',

    'The `draw_ellipse` interface draws the vector ellipse graphics.':
    '',

    '## Basic usage':
    '',

    'The `draw_ellipse` interface has the `x`\\, `y`\\, `width`\\, and `height` arguments. The `x` and `y` arguments are the ellipse center coordinates. The `width` and `height` arguments are the ellipse size. These sizes are twice the size of the radius.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=325,\n    stage_height=200,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\n\n# Set the cyan fill color and draw the ellipse.\nsprite.graphics.begin_fill(color=\'#0af\')\nsprite.graphics.draw_ellipse(x=125, y=100, width=150, height=100)\n\n# Set the only dotted-line style and draw the ellipse.\nsprite.graphics.begin_fill(color=\'\')\nsprite.graphics.line_style(\n    color=\'#fff\', thickness=3, dot_setting=ap.LineDotSetting(dot_size=3))\nsprite.graphics.draw_ellipse(x=200, y=100, width=150, height=100)\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_draw_ellipse_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_draw_ellipse_basic_usage/index.html" width="325" height="200"></iframe>':  # noqa
    '',

    '## Return value':
    '',

    'The return value of the `draw_ellipse` interface is the instance of the `Ellipse` class.\n\nIt has the basic interfaces (like the `x` or the `width` attributes) similar to the other graphics classes.\n\nThe following code example binds the click event handler. If you click the ellipse, the width and height become wider.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\n\ndef on_ellipse_click(\n        e: ap.MouseEvent[ap.Ellipse], options: dict) -> None:\n    """\n    The handler that the ellipse calls when clicked.\n\n    Parameters\n    ----------\n    e : MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    ellipse: ap.Ellipse = e.this\n    ellipse.width += 15\n    ellipse.height += 10\n\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=250,\n    stage_height=200,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\n\nsprite.graphics.begin_fill(color=\'#0af\')\nellipse: ap.Ellipse = sprite.graphics.draw_ellipse(\n    x=125, y=100, width=150, height=100)\nellipse.click(on_ellipse_click)\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_draw_ellipse_return_value/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_draw_ellipse_return_value/index.html" width="250" height="200"></iframe>':  # noqa
    '',

    '## draw_ellipse API':
    '',

    '<!-- Docstring: apysc._display.graphics.Graphics.draw_ellipse -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `draw_ellipse(self, x:Union[int, apysc._type.int.Int], y:Union[int, apysc._type.int.Int], width:Union[int, apysc._type.int.Int], height:Union[int, apysc._type.int.Int]) -> \'_ellipse.Ellipse\'`<hr>\n\n**[Interface summary]** Draw an ellipse vector graphic.<hr>\n\n**[Parameters]**\n\n- `x`: Int or int\n  - X-coordinate of the ellipse center.\n- `y`: Int or int\n  - Y-coordinate of the ellipse center.\n- `width`: Int or int\n  - Ellipse width.\n- `height`: Int or int\n  - Ellipse height.\n\n<hr>\n\n**[Returns]**\n\n- `ellipse`: Ellipse\n  - Created ellipse graphics instance.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> ellipse: ap.Ellipse = sprite.graphics.draw_ellipse(\n...     x=100, y=100, width=100, height=50)\n>>> ellipse.x\nInt(100)\n\n>>> ellipse.y\nInt(100)\n\n>>> ellipse.width\nInt(100)\n\n>>> ellipse.height\nInt(50)\n\n>>> ellipse.fill_color\nString(\'#00aaff\')\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: timer.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Timer':
    '',

    'This page explains the `Timer` class.':
    '',

    '## What is the Timer?':
    '',

    'The `Timer` class handles the timer\'s tick. You can call a handler at any intervals by it.':  # noqa
    '',

    '## Basic usage':
    '',

    'The `Timer` class requires arguments for the `handler` and `delay` (timer interval in milliseconds). And the `start` method starts that timer. A timer passes the `TimerEvent` instance and options arguments to a specified handler.\n\nThe following code sets the `Timer` when clicking the rectangle (`Sprite`):':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _RectOptions(TypedDict):\n    rectangle: ap.Rectangle\n\n\ndef on_sprite_click(\n        e: ap.MouseEvent[ap.Sprite], options: _RectOptions) -> None:\n    """\n    The Handler that the rectangle calls when clicked.\n\n    Parameters\n    ----------\n    e : MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    e.this.unbind_click_all()\n    timer: ap.Timer = ap.Timer(on_timer, delay=16.6, options=options)\n    timer.start()\n\n\ndef on_timer(e: ap.TimerEvent, options: _RectOptions) -> None:\n    """\n    The Handler a timer calls.\n\n    Parameters\n    ----------\n    e : TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = options[\'rectangle\']\n    rectangle.x += 1\n\n\nap.Stage(\n    stage_width=350, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\noptions: _RectOptions = {\'rectangle\': rectangle}\nsprite.click(on_sprite_click, options=options)\n\nap.save_overall_html(\n    dest_dir_path=\'timer_basic_usage/\')\n```':  # noqa
    '',

    'If you click the rectangle, the timer starts, and the Handler increases the rectangle x value.\n\n<iframe src="static/timer_basic_usage/index.html" width="350" height="150"></iframe>':  # noqa
    '',

    '## See also':
    '',

    '- [TimerEvent class](timer_event.md)\n- [Timer class delay setting](timer_delay.md)\n- [FPS enum](fps.md)\n- [Timer class repeat count setting](timer_repeat_count.md)\n- [Timer class start and stop interfaces](timer_start_and_stop.md)\n- [Timer class timer complete interface](timer_complete.md)\n- [Timer class reset interface](timer_reset.md)':  # noqa
    '',

    '## Timer constructor API':
    '',

    '<!-- Docstring: apysc._time.timer.Timer.__init__ -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `__init__(self, handler:Callable[[_ForwardRef(\'timer_event.TimerEvent\'), ~_O1], NoneType], delay:Union[int, float, apysc._type.number_value_interface.NumberValueInterface, apysc._time.fps.FPS], *, repeat_count:Union[int, apysc._type.int.Int]=0, options:Union[~_O1, NoneType]=None) -> None`<hr>\n\n**[Interface summary]** Timer class to handle function calling at regular intervals.<hr>\n\n**[Parameters]**\n\n- `handler`: _Handler\n  - A handler would be called at regular intervals.\n- `delay`: Int or int or Number or float or FPS\n  - A delay between each `Handler` calling in a millisecond or FPS value. If an `FPS` value is specified, this value becomes a millisecond calculated with that FPS value (e.g., if the `FPS_60` value is specified, then `delay` becomes 16.6666667).\n- `repeat_count`: Int or int\n  - Max count of a `Handler`\'s calling. A timer stops if the `Handler`\'s calling count has reached this value. If 0 is specified, then a timer loops forever.\n- `options`: dict or None, default None\n  - Optional arguments dictionary to pass a `Handler` callable.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> from typing_extensions import TypedDict\n>>> import apysc as ap\n>>> class RectOptions(TypedDict):\n...     rectangle: ap.Rectangle\n>>> def on_timer(e: ap.TimerEvent, options: RectOptions) -> None:\n...     rectangle: ap.Rectangle = options[\'rectangle\']\n...     rectangle.x += 1\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> options: RectOptions = {\'rectangle\': rectangle}\n>>> _ = ap.Timer(\n...     on_timer, delay=ap.FPS.FPS_60, options=options).start()\n```':  # noqa
    '',

    '<hr>\n\n**[References]**\n\n- [TimerEvent class document](https://simon-ritchie.github.io/apysc/timer_event.html)\n- [Timer class delay setting document](https://simon-ritchie.github.io/apysc/timer_delay.html)\n- [FPS enum document](https://simon-ritchie.github.io/apysc/fps.html)\n- [Timer class repeat_count setting](https://simon-ritchie.github.io/apysc/timer_repeat_count.html)\n- [About the handler options\' type document](https://simon-ritchie.github.io/apysc/about_handler_options_type.html)':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: display_object_x_and_y.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Display object x and y interfaces':
    '',

    'This page explains the `DisplayObject` class x and y property interfaces.':  # noqa
    '',

    '## What interfaces are these?':
    '',

    'The x and y properties change the `DisplayObject` instance 2-dimensional coordinates.':  # noqa
    '',

    '## Basic usage':
    '',

    'Each `DisplayObject` instance has the x and y properties and can get/set the value with it.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=0, y=0, width=50, height=50)\n\n# Update the x and y coordinates from 0 to 50.\nrectangle.x = ap.Int(50)\nrectangle.y = ap.Int(50)\n\nap.save_overall_html(\n    dest_dir_path=\'display_object_x_and_y_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/display_object_x_and_y_basic_usage/index.html" width="150" height="150"></iframe>':  # noqa
    '',

    '## Augmented assignment':
    '',

    'The x and y properties support the Augmented assignments, like the `+=`\\, `-=`\\, `/=`\\, and `*=` operators.\n\nThe following example appends 10-pixel to the y-coordinate when you click the rectangle.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\n\ndef on_click(e: ap.MouseEvent[ap.Rectangle], options: dict) -> None:\n    """\n    The handler that the rectangle calls when clicked.\n\n    Parameters\n    ----------\n    e : MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this\n    rectangle.y += 10\n\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nrectangle.click(on_click)\n\nap.save_overall_html(\n    dest_dir_path=\'display_object_x_and_y_augmented_assignment/\')\n```':  # noqa
    '',

    '<iframe src="static/display_object_x_and_y_augmented_assignment/index.html" width="150" height="150"></iframe>':  # noqa
    '',

    '## x property API':
    '',

    '<!-- Docstring: apysc._display.x_interface.XInterface.x -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get a x-coordinate.<hr>\n\n**[Returns]**\n\n- `x`: Int\n  - X-coordinate.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> rectangle.x = ap.Int(100)\n>>> rectangle.x\nInt(100)\n```':  # noqa
    '',

    '':
    '',

    '## y property API':
    '',

    '<!-- Docstring: apysc._display.y_interface.YInterface.y -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get a y-coordinate.<hr>\n\n**[Returns]**\n\n- `y`: Int\n  - Y-coordinate.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> rectangle.y = ap.Int(100)\n>>> rectangle.y\nInt(100)\n```':  # noqa
    '',

}
